# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Constants used to define the connections automatically added for each
PipelineTask by the execution system.
"""

from __future__ import annotations

__all__ = (
    "CONFIG_INIT_OUTPUT_CONNECTION_NAME",
    "CONFIG_INIT_OUTPUT_TEMPLATE",
    "CONFIG_INIT_OUTPUT_STORAGE_CLASS",
    "PACKAGES_INIT_OUTPUT_NAME",
    "PACKAGES_INIT_OUTPUT_STORAGE_CLASS",
    "LOG_OUTPUT_CONNECTION_NAME",
    "LOG_OUTPUT_TEMPLATE",
    "LOG_OUTPUT_STORAGE_CLASS",
    "METADATA_OUTPUT_CONNECTION_NAME",
    "METADATA_OUTPUT_TEMPLATE",
    "METADATA_OUTPUT_STORAGE_CLASS",
)


CONFIG_INIT_OUTPUT_CONNECTION_NAME: str = "_config"
"""Internal task-side name for the config dataset connection.
"""

CONFIG_INIT_OUTPUT_TEMPLATE: str = "{label}" + CONFIG_INIT_OUTPUT_CONNECTION_NAME
"""String template used to form the name for config init-output dataset
type names.
"""

CONFIG_INIT_OUTPUT_STORAGE_CLASS: str = "Config"
"""Name of the storage class for config init-output datasets.
"""

PACKAGES_INIT_OUTPUT_NAME: str = "packages"
"""Name of the global init-output dataset type that holds software versions.
"""

PACKAGES_INIT_OUTPUT_STORAGE_CLASS: str = "Packages"
"""Name of the storage class for the dataset type that holds software versions.
"""

LOG_OUTPUT_CONNECTION_NAME: str = "_log"
"""Internal task-side name for the log dataset connection.
"""

LOG_OUTPUT_TEMPLATE: str = "{label}" + LOG_OUTPUT_CONNECTION_NAME
"""String template used to form the name for log output dataset type names.
"""

LOG_OUTPUT_STORAGE_CLASS: str = "ButlerLogRecords"
"""Name of the storage class for log output datasets.
"""

METADATA_OUTPUT_CONNECTION_NAME: str = "_metadata"
"""Internal task-side name for the metadata dataset connection.
"""

METADATA_OUTPUT_TEMPLATE: str = "{label}" + METADATA_OUTPUT_CONNECTION_NAME
"""String template used to form the name for task metadata output dataset
type names.
"""

METADATA_OUTPUT_STORAGE_CLASS: str = "TaskMetadata"
"""Name of the storage class for task metadata output datasets.
"""



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the Instrument class.
"""

import datetime
import math
import unittest

from lsst.daf.butler import DataCoordinate, DimensionPacker, DimensionUniverse, RegistryConfig
from lsst.daf.butler.formatters.json import JsonFormatter
from lsst.daf.butler.registry.sql_registry import SqlRegistry
from lsst.pex.config import Config
from lsst.pipe.base import Instrument
from lsst.utils.introspection import get_full_type_name


class BaseDummyInstrument(Instrument):
    """Test instrument base class."""

    @classmethod
    def getName(cls):
        return "DummyInstrument"

    def register(self, registry, update=False):
        detector_max = 2
        visit_max = 10
        exposure_max = 8
        record = {
            "instrument": self.getName(),
            "class_name": get_full_type_name(DummyInstrument),
            "detector_max": detector_max,
            "visit_max": visit_max,
            "exposure_max": exposure_max,
        }
        with registry.transaction():
            registry.syncDimensionData("instrument", record, update=update)

    def getRawFormatter(self, dataId):
        return JsonFormatter


class DummyInstrument(BaseDummyInstrument):
    """Test instrument."""


class NotInstrument:
    """Not an instrument class at all."""

    def __init__(self, collection_prefix: str = ""):
        self.collection_prefix = collection_prefix


class BadInstrument(DummyInstrument):
    """Instrument with wrong class name."""

    raw_definition = ("raw2", ("instrument", "detector", "exposure"), "StructuredDataDict")

    @classmethod
    def getName(cls):
        return "BadInstrument"

    def register(self, registry, update=False):
        # Register a bad class name
        record = {
            "instrument": self.getName(),
            "class_name": "builtins.str",
            "detector_max": 1,
        }
        registry.syncDimensionData("instrument", record, update=update)


class UnimportableInstrument(DummyInstrument):
    """Instrument with class name that does not exist."""

    @classmethod
    def getName(cls):
        return "NoImportInstr"

    def register(self, registry, update=False):
        # Register a bad class name
        record = {
            "instrument": self.getName(),
            "class_name": "not.importable",
            "detector_max": 1,
        }
        registry.syncDimensionData("instrument", record, update=update)


class DimensionPackerTestConfig(Config):
    """Configuration for the dimension packer."""

    packer = Instrument.make_dimension_packer_config_field()


class InstrumentTestCase(unittest.TestCase):
    """Test for Instrument."""

    def setUp(self):
        self.instrument = DummyInstrument()
        self.name = "DummyInstrument"

    def test_basics(self):
        self.assertEqual(self.instrument.getName(), self.name)
        self.assertEqual(self.instrument.getRawFormatter({}), JsonFormatter)
        self.assertIsNone(DummyInstrument.raw_definition)
        raw = BadInstrument.raw_definition
        self.assertEqual(raw[2], "StructuredDataDict")

    def test_register(self):
        """Test that register() sets appropriate Dimensions."""
        registryConfig = RegistryConfig()
        registryConfig["db"] = "sqlite://"
        registry = SqlRegistry.createFromConfig(registryConfig)
        # Check that the registry starts out empty.
        self.instrument.importAll(registry)
        self.assertFalse(list(registry.queryDimensionRecords("instrument")))

        # Register and check again.
        self.instrument.register(registry)
        instruments = list(registry.queryDimensionRecords("instrument"))
        self.assertEqual(len(instruments), 1)
        self.assertEqual(instruments[0].name, self.name)
        self.assertEqual(instruments[0].detector_max, 2)
        self.assertIn("DummyInstrument", instruments[0].class_name)

        self.instrument.importAll(registry)
        from_registry = DummyInstrument.fromName("DummyInstrument", registry)
        self.assertIsInstance(from_registry, Instrument)
        with self.assertRaises(LookupError):
            Instrument.fromName("NotThrere", registry)

        # Register a bad instrument.
        BadInstrument().register(registry)
        with self.assertRaises(TypeError):
            Instrument.fromName("BadInstrument", registry)

        UnimportableInstrument().register(registry)
        with self.assertRaises(ImportError):
            Instrument.fromName("NoImportInstr", registry)

        # This should work even with the bad class name.
        self.instrument.importAll(registry)

        # Check that from_string falls back to fromName.
        from_string = DummyInstrument.from_string("DummyInstrument", registry)
        self.assertIsInstance(from_string, type(from_registry))

    def test_from_string(self):
        """Test Instrument.from_string works."""
        with self.assertRaises(RuntimeError):
            # No registry.
            DummyInstrument.from_string("DummyInstrument")

        with self.assertRaises(TypeError):
            # This will raise because collection_prefix is not understood.
            DummyInstrument.from_string("lsst.pipe.base.Task")

        with self.assertRaises(TypeError):
            # Not an instrument but does have collection_prefix.
            DummyInstrument.from_string(get_full_type_name(NotInstrument))

        with self.assertRaises(TypeError):
            # This will raise because BaseDummyInstrument is not a subclass
            # of DummyInstrument.
            DummyInstrument.from_string(get_full_type_name(BaseDummyInstrument))

        instrument = DummyInstrument.from_string(
            get_full_type_name(DummyInstrument), collection_prefix="test"
        )
        self.assertEqual(instrument.collection_prefix, "test")

    def test_defaults(self):
        self.assertEqual(self.instrument.makeDefaultRawIngestRunName(), "DummyInstrument/raw/all")
        self.assertEqual(
            self.instrument.makeUnboundedCalibrationRunName("a", "b"), "DummyInstrument/calib/a/b/unbounded"
        )
        self.assertEqual(
            self.instrument.makeCuratedCalibrationRunName("2018-05-04", "a"),
            "DummyInstrument/calib/a/curated/20180504T000000Z",
        )
        self.assertEqual(self.instrument.makeCalibrationCollectionName("c"), "DummyInstrument/calib/c")
        self.assertEqual(self.instrument.makeRefCatCollectionName(), "refcats")
        self.assertEqual(self.instrument.makeRefCatCollectionName("a"), "refcats/a")
        self.assertEqual(self.instrument.makeUmbrellaCollectionName(), "DummyInstrument/defaults")

        instrument = DummyInstrument(collection_prefix="Different")
        self.assertEqual(instrument.makeCollectionName("a"), "Different/a")
        self.assertEqual(self.instrument.makeCollectionName("a"), "DummyInstrument/a")

    def test_collection_timestamps(self):
        self.assertEqual(
            Instrument.formatCollectionTimestamp("2018-05-03"),
            "20180503T000000Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp("2018-05-03T14:32:16"),
            "20180503T143216Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp("20180503T143216Z"),
            "20180503T143216Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp(datetime.datetime(2018, 5, 3, 14, 32, 16)),
            "20180503T143216Z",
        )
        formattedNow = Instrument.makeCollectionTimestamp()
        self.assertIsInstance(formattedNow, str)
        datetimeThen1 = datetime.datetime.strptime(formattedNow, "%Y%m%dT%H%M%S%z")
        self.assertEqual(datetimeThen1.tzinfo, datetime.timezone.utc)

        with self.assertRaises(TypeError):
            Instrument.formatCollectionTimestamp(0)

    def test_dimension_packer_config_defaults(self):
        """Test the ObservationDimensionPacker class and the Instrument-based
        systems for constructing it, in the case where the Instrument-defined
        default is used.
        """
        registry_config = RegistryConfig()
        registry_config["db"] = "sqlite://"
        registry = SqlRegistry.createFromConfig(registry_config)
        self.instrument.register(registry)
        config = DimensionPackerTestConfig()
        instrument_data_id = registry.expandDataId(instrument=self.name)
        record = instrument_data_id.records["instrument"]
        self.check_dimension_packers(
            registry.dimensions,
            # Test just one packer-construction signature here as that saves us
            # from having to insert dimension records for visits, exposures,
            # and detectors for this test.
            # Note that we don't need to pass any more than the instrument in
            # the data ID yet, because we're just constructing packers, not
            # calling their pack method.
            visit_packers=[
                config.packer.apply(instrument_data_id, is_exposure=False),
                Instrument.make_default_dimension_packer(instrument_data_id, is_exposure=False),
            ],
            exposure_packers=[
                config.packer.apply(instrument_data_id, is_exposure=True),
                Instrument.make_default_dimension_packer(instrument_data_id, is_exposure=True),
            ],
            n_detectors=record.detector_max,
            n_visits=record.visit_max,
            n_exposures=record.exposure_max,
        )

    def test_dimension_packer_config_override(self):
        """Test the ObservationDimensionPacker class and the Instrument-based
        systems for constructing it, in the case where configuration overrides
        the Instrument's default.
        """
        registry_config = RegistryConfig()
        registry_config["db"] = "sqlite://"
        registry = SqlRegistry.createFromConfig(registry_config)
        # Intentionally do not register instrument or insert any other
        # dimension records to ensure we don't need them in this mode.
        config = DimensionPackerTestConfig()
        config.packer["observation"].n_observations = 16
        config.packer["observation"].n_detectors = 4
        config.packer.name = "observation"
        instrument_data_id = DataCoordinate.standardize(instrument=self.name, universe=registry.dimensions)
        full_data_id = DataCoordinate.standardize(instrument_data_id, detector=1, visit=3, exposure=7)
        visit_data_id = full_data_id.subset(full_data_id.universe.conform(["visit", "detector"]))
        exposure_data_id = full_data_id.subset(full_data_id.universe.conform(["exposure", "detector"]))
        self.check_dimension_packers(
            registry.dimensions,
            # Note that we don't need to pass any more than the instrument in
            # the data ID yet, because we're just constructing packers, not
            # calling their pack method.
            visit_packers=[
                config.packer.apply(visit_data_id),
                config.packer.apply(full_data_id),
                config.packer.apply(instrument_data_id, is_exposure=False),
            ],
            exposure_packers=[
                config.packer.apply(instrument_data_id, is_exposure=True),
                config.packer.apply(exposure_data_id),
            ],
            n_detectors=config.packer["observation"].n_detectors,
            n_visits=config.packer["observation"].n_observations,
            n_exposures=config.packer["observation"].n_observations,
        )

    def check_dimension_packers(
        self,
        universe: DimensionUniverse,
        visit_packers: list[DimensionPacker],
        exposure_packers: list[DimensionPacker],
        n_detectors: int,
        n_visits: int,
        n_exposures: int,
    ) -> None:
        """Test the behavior of one or more dimension packers constructed by
        an instrument.

        Parameters
        ----------
        universe : `lsst.daf.butler.DimensionUniverse`
            Data model for butler data IDs.
        visit_packers : `list` [ `DimensionPacker` ]
            Packers with ``{visit, detector}`` dimensions to test.
        exposure_packers : `list` [ `DimensionPacker` ]
            Packers with ``{exposure, detector}`` dimensions to test.
        n_detectors : `int`
            Number of detectors all packers have been configured to reserve
            space for.
        n_visits : `int`
            Number of visits all packers in ``visit_packers`` have been
            configured to reserve space for.
        n_exposures : `int`
            Number of exposures all packers in ``exposure_packers`` have been
            configured to reserve space for.
        """
        full_data_id = DataCoordinate.standardize(
            instrument=self.name, detector=1, visit=3, exposure=7, universe=universe
        )
        visit_data_id = full_data_id.subset(full_data_id.universe.conform(["visit", "detector"]))
        exposure_data_id = full_data_id.subset(full_data_id.universe.conform(["exposure", "detector"]))
        for n, packer in enumerate(visit_packers):
            with self.subTest(n=n):
                packed_id, max_bits = packer.pack(full_data_id, returnMaxBits=True)
                self.assertEqual(packed_id, full_data_id["detector"] + full_data_id["visit"] * n_detectors)
                self.assertEqual(max_bits, math.ceil(math.log2(n_detectors * n_visits)))
                self.assertEqual(visit_data_id, packer.unpack(packed_id))
                with self.assertRaisesRegex(ValueError, f"Detector ID {n_detectors} is out of bounds"):
                    packer.pack(instrument=self.name, detector=n_detectors, visit=3)
                with self.assertRaisesRegex(ValueError, f"Visit ID {n_visits} is out of bounds"):
                    packer.pack(instrument=self.name, detector=1, visit=n_visits)
        for n, packer in enumerate(exposure_packers):
            with self.subTest(n=n):
                packed_id, max_bits = packer.pack(full_data_id, returnMaxBits=True)
                self.assertEqual(packed_id, full_data_id["detector"] + full_data_id["exposure"] * n_detectors)
                self.assertEqual(max_bits, math.ceil(math.log2(n_detectors * n_exposures)))
                self.assertEqual(exposure_data_id, packer.unpack(packed_id))
                with self.assertRaisesRegex(ValueError, f"Detector ID {n_detectors} is out of bounds"):
                    packer.pack(instrument=self.name, detector=n_detectors, exposure=3)
                with self.assertRaisesRegex(ValueError, f"Exposure ID {n_exposures} is out of bounds"):
                    packer.pack(instrument=self.name, detector=1, exposure=n_exposures)


if __name__ == "__main__":
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = (
    "MergedNodeKey",
    "merge_graph_input_trees",
    "merge_graph_output_trees",
    "merge_graph_intermediates",
)

import dataclasses
import hashlib
from collections import defaultdict
from functools import cached_property
from typing import Any, Iterable, TypeVar

import networkx
import networkx.algorithms.dag
import networkx.algorithms.tree
from lsst.daf.butler import DimensionGroup

from .._nodes import NodeKey, NodeType
from ._options import NodeAttributeOptions

_P = TypeVar("_P")
_C = TypeVar("_C")


class MergedNodeKey(frozenset[NodeKey]):
    """A key for NetworkX graph nodes that represent multiple similar tasks
    or dataset types that have been merged to simplify graph visualization.
    """

    def __str__(self) -> str:
        members = [str(k) for k in self]
        members.sort(reverse=True)
        return ", ".join(members)

    @property
    def node_type(self) -> NodeType:
        """Enum value for whether this is a task, task initialization, or
        dataset type node.
        """
        return next(iter(self)).node_type

    @cached_property
    def node_id(self) -> str:
        """A unique string representation of the merged node key."""
        hasher = hashlib.blake2b(digest_size=4)
        for member in sorted(self):
            hasher.update(str(member).encode())
        return f"{hasher.digest().hex()}:{self.node_type.value}"


def merge_graph_input_trees(
    xgraph: networkx.DiGraph | networkx.MultiDiGraph, options: NodeAttributeOptions, depth: int
) -> None:
    """Merge trees of overall-input dataset type nodes and/or
    beginning-of-pipeline task nodes that have similar properties and the same
    structure.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        Graph to be processed; modified in-place.
    options : `NodeAttributeOptions`
        Properties of nodes that should be considered when determining whether
        they are similar enough to be merged.  Only the truthiness of
        attributes is considered (e.g. ``options.dimensions == 'full'`` and
        ``options.dimensions == 'concise'`` are both interpreted to mean "only
        merge trees with the same dimensions").  This is typically the same set
        of options that controls whether to display these attributes in the
        graph visualization.
    depth : `int`
        How many nodes to traverse from the beginning of the graph before
        terminating the merging algorithm.
    """
    groups = _make_tree_merge_groups(xgraph, options, depth)
    _apply_tree_merges(xgraph, groups)


def merge_graph_output_trees(
    xgraph: networkx.DiGraph | networkx.MultiDiGraph, options: NodeAttributeOptions, depth: int
) -> None:
    """Merge trees of overall-output dataset type nodes and/or
    end-of-pipeline task nodes that have similar properties and the same
    structure.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        Graph to be processed; modified in-place.
    options : `NodeAttributeOptions`
        Properties of nodes that should be considered when determining whether
        they are similar enough to be merged.  Only the truthiness of
        attributes is considered (e.g. ``options.dimensions == 'full'`` and
        ``options.dimensions == 'concise'`` are both interpreted to mean "only
        merge trees with the same dimensions").  This is typically the same set
        of options that controls whether to display these attributes in the
        graph visualization.
    depth : `int`
        How many nodes to traverse from the beginning of the graph before
        terminating the merging algorithm.
    """
    groups = _make_tree_merge_groups(xgraph.reverse(copy=False), options, depth)
    _apply_tree_merges(xgraph, groups)


def merge_graph_intermediates(
    xgraph: networkx.DiGraph | networkx.MultiDiGraph, options: NodeAttributeOptions
) -> None:
    """Merge parallel interior nodes of a graph with similar properties.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        Graph to be processed; modified in-place.
    options : `NodeAttributeOptions`
        Properties of nodes that should be considered when determining whether
        they are similar enough to be merged.  Only the truthiness of
        attributes is considered (e.g. ``options.dimensions == 'full'`` and
        ``options.dimensions == 'concise'`` are both interpreted to mean "only
        merge trees with the same dimensions").  This is typically the same set
        of options that controls whether to display these attributes in the
        graph visualization.

    Notes
    -----
    "Parallel" nodes here are nodes that have the exact same predecessor and
    successors.
    """
    groups: dict[_MergeKey, set[NodeKey]] = defaultdict(set)
    for node, state in xgraph.nodes.items():
        merge_key = _MergeKey.from_node_state(
            state,
            xgraph.predecessors(node),
            xgraph.successors(node),
            options,
        )
        if merge_key.parents and merge_key.children:
            groups[merge_key].add(node)
    replacements: dict[NodeKey, MergedNodeKey] = {}
    for merge_key, members in groups.items():
        if len(members) < 2:
            continue
        new_node_key = MergedNodeKey(frozenset(members))
        xgraph.add_node(
            new_node_key,
            storage_class_name=merge_key.storage_class_name,
            task_class_name=merge_key.task_class_name,
            dimensions=merge_key.dimensions,
        )
        for parent in merge_key.parents:
            xgraph.add_edge(replacements.get(parent, parent), new_node_key)
        for child in merge_key.children:
            xgraph.add_edge(new_node_key, replacements.get(child, child))
        for member in members:
            replacements[member] = new_node_key
        xgraph.remove_nodes_from(members)


@dataclasses.dataclass(frozen=True)
class _MergeKey:
    """A helper class for merge algorithms that is used as a dictionary key
    when grouping nodes that may be merged by their attributes.
    """

    parents: frozenset[Any]
    """Nodes of the original graph that are successors or predecessors of
    the nodes being considered for merging.
    """

    dimensions: DimensionGroup | None
    """Dimensions of the nodes being considered for merging, or `None` if
    dimensions are not included in the similarity criteria.
    """

    storage_class_name: str | None
    """Storage class of the nodes being considered for merging, or `None` if
    storage classes are not included in the similarity criteria or this is a
    task or task initialization node group.
    """

    task_class_name: str | None
    """Name of the task class for the nodes being considered for merging, or
    `None` if task classes are not included in the similarity criteria or
    this is a dataset type node group.
    """

    children: frozenset[Any]
    """Nodes that are predecessors or successors (the opposite of ``parents``
    of the nodes being considered for merging.

    In the `merge_graph_intermediates` algorithm, these are regular unmerged
    nodes.  In the `merge_graph_input_trees` or `merge_graph_output_trees`
    algorithms, these are more `_MergeKey` instances, representing
    already-processed trees.
    """

    @classmethod
    def from_node_state(
        cls,
        state: dict[str, Any],
        parents: Iterable[_P],
        children: Iterable[_C],
        options: NodeAttributeOptions,
    ) -> _MergeKey:
        """Construct from a NetworkX node attribute state dictionary.

        Parameters
        ----------
        state : `dict`
            Dictionary used to hold NetworkX node attributes.
        parents : `~collections.abc.Iterable` [ `NodeKey` ]
            Predecessor or successor nodes (depending on the orientation of
            the algorithm).
        children : ~collections.abc.Iterable`
            Successor or predecessor nodes (depending on the orientation of
            the algorithm).
        options : `NodeAttributeOptions`
            Options for which node attributes to include in the new key.
        """
        return cls(
            parents=frozenset(parents),
            dimensions=state.get("dimensions"),
            storage_class_name=(state.get("storage_class_name") if options.storage_classes else None),
            task_class_name=(state.get("task_class_name") if options.task_classes else None),
            children=frozenset(children),
        )


def _make_tree_merge_groups(
    xgraph: networkx.DiGraph | networkx.MultiDiGraph,
    options: NodeAttributeOptions,
    depth: int,
) -> list[dict[_MergeKey, set[NodeKey]]]:
    """First-stage implementation of `merge_graph_input_trees` and
    (when run on the reversed graph) `merge_graph_output_trees`.
    """
    # Our goal is to obtain mappings that groups trees of nodes by the
    # attributes in a _TreeMergeKey.  The nested dictionaries are the root of a
    # tree and the nodes under that root, recursively (but not including the
    # root).  We nest these mappings inside a list, which each mapping
    # corresponding to a different depth for the trees it represents.  We start
    # with a special empty dict for "0-depth trees", since that makes
    # result[depth] valid and hence off-by-one errors less likely.
    result: list[dict[_MergeKey, set[NodeKey]]] = [{}]
    if depth == 0:
        return result
    # We start with the nodes that have no predecessors in the graph.
    # Ignore for now the fact that the 'current_candidates' data structure
    # we process is actually a dict that associates each of those nodes
    # with an empty dict.  All of these initial nodes are valid trees,
    # since they're just single nodes.
    first_generation = next(networkx.algorithms.dag.topological_generations(xgraph))
    current_candidates: dict[NodeKey, dict[NodeKey, _MergeKey]] = dict.fromkeys(first_generation, {})
    # Set up an outer loop over tree depth; we'll construct a new set of
    # candidates at each iteration.
    while current_candidates:
        # As we go, we'll remember nodes that have just one predecessor, as
        # those predecessors might be the roots of slightly taller trees.
        # We store the successors and their merge keys under them.
        next_candidates: dict[NodeKey, dict[NodeKey, _MergeKey]] = defaultdict(dict)
        # We also want to track the nodes the level up that are not trees
        # because some node has both them and some other node as a
        # predecessor.
        nontrees: set[NodeKey] = set()
        # Make a dictionary for the results at this depth, then start the
        # inner iteration over candidates and (after the first iteration)
        # their children.
        result_for_depth: dict[_MergeKey, set[NodeKey]] = defaultdict(set)
        for node, children in current_candidates.items():
            # Make a _TreeMergeKey for this node and add it to the results for
            # this depth.  Two nodes with the same _TreeMergeKey are roots of
            # isomorphic trees that have the same predecessor(s), and can be
            # merged (with isomorphism defined as both both structure and
            # whatever comparisons are in 'options').
            merge_key = _MergeKey.from_node_state(
                xgraph.nodes[node], xgraph.successors(node), children.values(), options
            )
            result_for_depth[merge_key].add(node)
            if len(result) <= depth:
                # See if this node's successor might be the root of a
                # larger tree.
                if len(merge_key.parents) == 1:
                    (parent,) = merge_key.parents
                    next_candidates[parent][node] = dataclasses.replace(merge_key, parents=frozenset())
                else:
                    nontrees.update(merge_key.parents)
        # Append the results for this depth.
        result.append(result_for_depth)
        # Trim out candidates that aren't trees after all.
        for nontree_node in nontrees & next_candidates.keys():
            del next_candidates[nontree_node]
        current_candidates = next_candidates
    return result


def _apply_tree_merges(
    xgraph: networkx.DiGraph | networkx.MultiDiGraph,
    groups: list[dict[_MergeKey, set[NodeKey]]],
) -> None:
    """Second-stage implementation of `merge_graph_input_trees` and
    `merge_graph_output_trees`.
    """
    replacements: dict[NodeKey, MergedNodeKey] = {}
    for group in reversed(groups):
        new_group: dict[_MergeKey, set[NodeKey]] = defaultdict(set)
        for merge_key, members in group.items():
            if merge_key.parents & replacements.keys():
                replaced_parents = frozenset(replacements.get(p, p) for p in merge_key.parents)
                new_group[dataclasses.replace(merge_key, parents=replaced_parents)].update(members)
            else:
                new_group[merge_key].update(members)
        for merge_key, members in new_group.items():
            if len(members) < 2:
                continue
            new_node_key = MergedNodeKey(frozenset(members))
            new_edges: set[tuple[NodeKey | MergedNodeKey, NodeKey | MergedNodeKey]] = set()
            for member_key in members:
                replacements[member_key] = new_node_key
                new_edges.update(
                    (replacements.get(a, a), replacements.get(b, b)) for a, b in xgraph.in_edges(member_key)
                )
                new_edges.update(
                    (replacements.get(a, a), replacements.get(b, b)) for a, b in xgraph.out_edges(member_key)
                )
            xgraph.add_node(
                new_node_key,
                storage_class_name=merge_key.storage_class_name,
                task_class_name=merge_key.task_class_name,
                dimensions=merge_key.dimensions,
            )
            xgraph.add_edges_from(new_edges)
    xgraph.remove_nodes_from(replacements.keys())



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The standard, general-purpose implementation of the QuantumGraph-generation
algorithm.
"""

from __future__ import annotations

__all__ = ("AllDimensionsQuantumGraphBuilder", "DatasetQueryConstraintVariant")

import dataclasses
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from io import StringIO
from typing import TYPE_CHECKING, Any, final

from lsst.daf.butler.registry import MissingDatasetTypeError
from lsst.utils.timer import timeMethod

from ._datasetQueryConstraints import DatasetQueryConstraintVariant
from .quantum_graph_builder import (
    DatasetKey,
    PrerequisiteDatasetKey,
    QuantumGraphBuilder,
    QuantumGraphBuilderError,
    QuantumGraphSkeleton,
    QuantumKey,
)

if TYPE_CHECKING:
    from lsst.daf.butler import Butler, DimensionGroup
    from lsst.daf.butler.registry.queries import DataCoordinateQueryResults
    from lsst.utils.logging import LsstLogAdapter

    from .pipeline_graph import DatasetTypeNode, PipelineGraph, TaskNode


@final
class AllDimensionsQuantumGraphBuilder(QuantumGraphBuilder):
    """An implementation of `QuantumGraphBuilder` that uses a single large
    query for data IDs covering all dimensions in the pipeline.

    Parameters
    ----------
    pipeline_graph : `.pipeline_graph.PipelineGraph`
        Pipeline to build a `QuantumGraph` from, as a graph.  Will be resolved
        in-place with the given butler (any existing resolution is ignored).
    butler : `lsst.daf.butler.Butler`
        Client for the data repository.  Should be read-only.
    where : `str`, optional
        Butler expression language constraint to apply to all data IDs.
    dataset_query_constraint : `DatasetQueryConstraintVariant`, optional
        Specification of which overall-input datasets should be used to
        constrain the initial data ID queries.  Not including an important
        constraint can result in catastrophically large query results that take
        too long to process, while including too many makes the query much more
        complex, increasing the chances that the database will choose a bad
        (sometimes catastrophically bad) query plan.
    bind : `~collections.abc.Mapping`, optional
        Variable substitutions for the ``where`` expression.
    **kwargs
        Additional keyword arguments forwarded to `QuantumGraphBuilder`.

    Notes
    -----
    This is a general-purpose algorithm that delegates the problem of
    determining which "end" of the pipeline is more constrained (beginning by
    input collection contents vs. end by the ``where`` string) to the database
    query planner, which *usually* does a good job.

    This algorithm suffers from a serious limitation, which we refer to as the
    "tract slicing" problem from its most common variant: the ``where`` string
    and general data ID intersection rules apply to *all* data IDs in the
    graph.  For example, if a ``tract`` constraint is present in the ``where``
    string or an overall-input dataset, then it is impossible for any data ID
    that does not overlap that tract to be present anywhere in the pipeline,
    such as a ``{visit, detector}`` combination where the ``visit`` overlaps
    the ``tract`` even if the ``detector`` does not.
    """

    def __init__(
        self,
        pipeline_graph: PipelineGraph,
        butler: Butler,
        *,
        where: str = "",
        dataset_query_constraint: DatasetQueryConstraintVariant = DatasetQueryConstraintVariant.ALL,
        bind: Mapping[str, Any] | None = None,
        **kwargs: Any,
    ):
        super().__init__(pipeline_graph, butler, **kwargs)
        self.where = where
        self.dataset_query_constraint = dataset_query_constraint
        self.bind = bind

    @timeMethod
    def process_subgraph(self, subgraph: PipelineGraph) -> QuantumGraphSkeleton:
        # Docstring inherited.
        # There is some chance that the dimension query for one subgraph would
        # be the same as or a dimension-subset of another.  This is an
        # optimization opportunity we're not currently taking advantage of.
        with _AllDimensionsQuery.from_builder(self, subgraph) as query:
            skeleton = self._make_subgraph_skeleton(query)
            self._find_followup_datasets(query, skeleton)
        return skeleton

    @timeMethod
    def _make_subgraph_skeleton(self, query: _AllDimensionsQuery) -> QuantumGraphSkeleton:
        """Build a `QuantumGraphSkeleton` by iterating over the result rows
        of the initial data ID query.

        Parameters
        ----------
        query : `_AllDimensionsQuery`
            Object representing the full-pipeline data ID query.

        Returns
        -------
        skeleton : `QuantumGraphSkeleton`
            Preliminary quantum graph.
        """
        # First we make containers of empty-dimensions quantum and dataset
        # keys, and add those to the skelton, since empty data IDs are
        # logically subsets of any data ID. We'll copy those to initialize the
        # containers of keys for each result row.  We don't ever explicitly add
        # nodes to the skeleton for these, and that's okay because networkx
        # adds nodes implicitly when an edge to that node is added, and we
        # don't want to add nodes for init datasets here.
        skeleton = QuantumGraphSkeleton(query.subgraph.tasks)
        empty_dimensions_dataset_keys = {}
        for dataset_type_name in query.empty_dimensions_dataset_types.keys():
            empty_dimensions_dataset_keys[dataset_type_name] = skeleton.add_dataset_node(
                dataset_type_name, self.empty_data_id
            )
        empty_dimensions_quantum_keys = []
        for task_label in query.empty_dimensions_tasks.keys():
            empty_dimensions_quantum_keys.append(skeleton.add_quantum_node(task_label, self.empty_data_id))
        self.log.info("Iterating over query results to associate quanta with datasets.")
        # Iterate over query results, populating data IDs for datasets and
        # quanta and then connecting them to each other. This is the slowest
        # client-side part of QG generation, and it's often the slowest part
        # overall, so inside this loop is where it's really critical to avoid
        # expensive things, especially in the nested loops.
        n_rows = 0
        for common_data_id in query.common_data_ids:
            # Create a data ID for each set of dimensions used by one or more
            # tasks or dataset types, and use that to record all quanta and
            # dataset data IDs for this row.
            dataset_keys_for_row: dict[str, DatasetKey] = empty_dimensions_dataset_keys.copy()
            quantum_keys_for_row: list[QuantumKey] = empty_dimensions_quantum_keys.copy()
            for dimensions, (task_nodes, dataset_type_nodes) in query.grouped_by_dimensions.items():
                data_id = common_data_id.subset(dimensions)
                for dataset_type_name in dataset_type_nodes.keys():
                    dataset_keys_for_row[dataset_type_name] = skeleton.add_dataset_node(
                        dataset_type_name, data_id
                    )
                for task_label in task_nodes.keys():
                    quantum_keys_for_row.append(skeleton.add_quantum_node(task_label, data_id))
            # Whether these quanta are new or existing, we can now associate
            # the dataset data IDs for this row with them.  The fact that a
            # quantum data ID and a dataset data ID both came from the same
            # result row is what tells us they should be associated.  Many of
            # these associates will be duplicates (because another query row
            # that differed from this one only in irrelevant dimensions already
            # added them), and our use of sets should take care of that.
            for quantum_key in quantum_keys_for_row:
                for read_edge in self._pipeline_graph.tasks[quantum_key.task_label].inputs.values():
                    skeleton.add_input_edge(
                        quantum_key, dataset_keys_for_row[read_edge.parent_dataset_type_name]
                    )
                for write_edge in self._pipeline_graph.tasks[quantum_key.task_label].iter_all_outputs():
                    skeleton.add_output_edge(
                        quantum_key, dataset_keys_for_row[write_edge.parent_dataset_type_name]
                    )
            n_rows += 1
        if n_rows == 0:
            query.log_failure(self.log)
        else:
            n_quanta = sum(len(skeleton.get_quanta(task_label)) for task_label in query.subgraph.tasks)
            self.log.info(
                "Initial bipartite graph has %d quanta, %d dataset nodes, and %d edges from %d query row(s).",
                n_quanta,
                skeleton.n_nodes - n_quanta,
                skeleton.n_edges,
                n_rows,
            )
        return skeleton

    @timeMethod
    def _find_followup_datasets(self, query: _AllDimensionsQuery, skeleton: QuantumGraphSkeleton) -> None:
        """Populate `existing_datasets` by performing follow-up queries joined
        to column-subsets of the initial data ID query.

        Parameters
        ----------
        query : `_AllDimensionsQuery`
            Object representing the full-pipeline data ID query.
        """
        for dimensions, (tasks_in_group, dataset_types_in_group) in query.grouped_by_dimensions.items():
            data_ids = query.common_data_ids.subset(dimensions, unique=True)
            # Iterate over regular input/output dataset type nodes with these
            # dimensions to find those datasets using straightforward followup
            # queries.
            for dataset_type_node in dataset_types_in_group.values():
                if dataset_type_node.name in query.overall_inputs:
                    # Dataset type is an overall input; we always need to try
                    # to find these.
                    count = 0
                    try:
                        for ref in data_ids.findDatasets(dataset_type_node.name, self.input_collections):
                            self.existing_datasets.inputs[
                                DatasetKey(dataset_type_node.name, ref.dataId.required_values)
                            ] = ref
                            count += 1
                    except MissingDatasetTypeError:
                        pass
                    self.log.verbose(
                        "Found %d overall-input dataset(s) of type %r.", count, dataset_type_node.name
                    )
                    continue
                if self.skip_existing_in:
                    # Dataset type is an intermediate or output; need to find
                    # these if only they're from previously executed quanta
                    # that we might skip...
                    count = 0
                    try:
                        for ref in data_ids.findDatasets(dataset_type_node.name, self.skip_existing_in):
                            key = DatasetKey(dataset_type_node.name, ref.dataId.required_values)
                            self.existing_datasets.outputs_for_skip[key] = ref
                            count += 1
                            if ref.run == self.output_run:
                                self.existing_datasets.outputs_in_the_way[key] = ref
                    except MissingDatasetTypeError:
                        pass
                    self.log.verbose(
                        "Found %d output dataset(s) of type %r in %s.",
                        count,
                        dataset_type_node.name,
                        self.skip_existing_in,
                    )
                if self.output_run_exists and not self.skip_existing_starts_with_output_run:
                    # ...or if they're in the way and would need to be
                    # clobbered (and we haven't already found them in the
                    # previous block).
                    count = 0
                    try:
                        for ref in data_ids.findDatasets(dataset_type_node.name, [self.output_run]):
                            self.existing_datasets.outputs_in_the_way[
                                DatasetKey(dataset_type_node.name, ref.dataId.required_values)
                            ] = ref
                            count += 1
                    except MissingDatasetTypeError:
                        pass
                    self.log.verbose(
                        "Found %d output dataset(s) of type %r in %s.",
                        count,
                        dataset_type_node.name,
                        self.output_run,
                    )
            del dataset_type_node
            # Iterate over tasks with these dimensions to perform follow-up
            # queries for prerequisite inputs, which may have dimensions that
            # were not in ``common_data_ids`` and/or require temporal joins to
            # calibration validity ranges.
            for task_node in tasks_in_group.values():
                task_prerequisite_info = self.prerequisite_info[task_node.label]
                for connection_name, finder in list(task_prerequisite_info.finders.items()):
                    if finder.lookup_function is not None:
                        self.log.verbose(
                            "Deferring prerequisite input %r of task %r to per-quantum processing "
                            "(lookup function provided).",
                            finder.dataset_type_node.name,
                            task_node.label,
                        )
                        continue
                    # We also fall back to the base class if there is a
                    # nontrivial spatial or temporal join in the lookup.
                    if finder.dataset_skypix or finder.dataset_other_spatial:
                        if task_prerequisite_info.bounds.spatial_connections:
                            self.log.verbose(
                                "Deferring prerequisite input %r of task %r to per-quantum processing "
                                "(for spatial-bounds-connections handling).",
                                finder.dataset_type_node.name,
                                task_node.label,
                            )
                            continue
                        if not task_node.dimensions.spatial:
                            self.log.verbose(
                                "Deferring prerequisite input %r of task %r to per-quantum processing "
                                "(dataset has spatial data IDs, but task does not).",
                                finder.dataset_type_node.name,
                                task_node.label,
                            )
                            continue
                    if finder.dataset_has_timespan:
                        if task_prerequisite_info.bounds.spatial_connections:
                            self.log.verbose(
                                "Deferring prerequisite input %r of task %r to per-quantum processing "
                                "(for temporal-bounds-connections handling).",
                                finder.dataset_type_node.name,
                                task_node.label,
                            )
                            continue
                        if not task_node.dimensions.temporal:
                            self.log.verbose(
                                "Deferring prerequisite input %r of task %r to per-quantum processing "
                                "(dataset has temporal data IDs, but task does not).",
                                finder.dataset_type_node.name,
                                task_node.label,
                            )
                            continue
                    # We have a simple case where we can do a single query
                    # that joins the query we already have for the task data
                    # IDs to the datasets we're looking for.
                    count = 0
                    try:
                        query_results = data_ids.findRelatedDatasets(
                            finder.dataset_type_node.dataset_type, self.input_collections
                        )
                    except MissingDatasetTypeError:
                        query_results = []
                    for data_id, ref in query_results:
                        dataset_key = PrerequisiteDatasetKey(finder.dataset_type_node.name, ref.id.bytes)
                        quantum_key = QuantumKey(task_node.label, data_id.required_values)
                        # The column-subset operation used to make `data_ids`
                        # from `common_data_ids` can strip away post-query
                        # filtering; e.g. if we starts with a {visit, patch}
                        # query but subset down to just {visit}, we can't keep
                        # the patch.region column we need for that filtering.
                        # This means we can get some data IDs that weren't in
                        # the original query (e.g. visits that don't overlap
                        # the same patch, but do overlap the some common skypix
                        # ID).  We don't want to add quanta with those data ID
                        # here, which is why we pass
                        # ignore_unrecognized_quanta=True here.
                        if skeleton.add_input_edge(quantum_key, dataset_key, ignore_unrecognized_quanta=True):
                            self.existing_datasets.inputs[dataset_key] = ref
                        count += 1
                    # Remove this finder from the mapping so the base class
                    # knows it doesn't have to look for these prerequisites.
                    del task_prerequisite_info.finders[connection_name]
                    self.log.verbose(
                        "Added %d prerequisite input edge(s) from dataset type %r to task %r.",
                        count,
                        finder.dataset_type_node.name,
                        task_node.label,
                    )


@dataclasses.dataclass(eq=False, repr=False)
class _AllDimensionsQuery:
    """A helper class for `AllDimensionsQuantumGraphBuilder` that holds all
    per-subgraph state.

    This object should always be constructed by `from_builder`, which returns
    an instance wrapped with a context manager. This controls the lifetime of
    the temporary table referenced by `common_data_ids`.
    """

    subgraph: PipelineGraph
    """Graph of this subset of the pipeline."""

    grouped_by_dimensions: dict[DimensionGroup, tuple[dict[str, TaskNode], dict[str, DatasetTypeNode]]] = (
        dataclasses.field(default_factory=dict)
    )
    """The tasks and dataset types of this subset of the pipeline, grouped
    by their dimensions.

    The tasks and dataset types with empty dimensions are not included; they're
    in other attributes since they are usually used differently.  Prerequisite
    dataset types are also not included.
    """

    empty_dimensions_tasks: dict[str, TaskNode] = dataclasses.field(default_factory=dict)
    """The tasks of this subset of this pipeline that have empty dimensions."""

    empty_dimensions_dataset_types: dict[str, DatasetTypeNode] = dataclasses.field(default_factory=dict)
    """The dataset types of this subset of this pipeline that have empty
    dimensions.

    Prerequisite dataset types are not included.
    """

    overall_inputs: dict[str, DatasetTypeNode] = dataclasses.field(default_factory=dict)
    """Pipeline graph nodes for all non-prerequisite, non-init overall-input
    dataset types for this subset of the pipeline.
    """

    query_args: dict[str, Any] = dataclasses.field(default_factory=dict)
    """All keyword arguments passed to `lsst.daf.butler.Registry.queryDataIds`.
    """

    common_data_ids: DataCoordinateQueryResults = dataclasses.field(init=False)
    """Results of the materialized initial data ID query."""

    @classmethod
    @contextmanager
    def from_builder(
        cls, builder: AllDimensionsQuantumGraphBuilder, subgraph: PipelineGraph
    ) -> Iterator[_AllDimensionsQuery]:
        """Construct and run the query, returning an instance guarded by
        a context manager.

        Parameters
        ----------
        builder : `AllDimensionsQuantumGraphBuilder`
            Builder object this helper is associated with.
        subgraph : `pipeline_graph.PipelineGraph`
            Subset of the pipeline being processed.

        Returns
        -------
        context : `AbstractContextManager` [ `_AllDimensionsQuery` ]
            An instance of this class, inside a context manager that manages
            the lifetime of its temporary database table.
        """
        result = cls(subgraph)
        builder.log.debug("Analyzing subgraph dimensions and overall-inputs.")
        result.grouped_by_dimensions = result.subgraph.group_by_dimensions()
        (
            result.empty_dimensions_tasks,
            result.empty_dimensions_dataset_types,
        ) = result.grouped_by_dimensions.pop(builder.universe.empty)
        result.overall_inputs = {
            name: node  # type: ignore
            for name, node in result.subgraph.iter_overall_inputs()
            if not node.is_prerequisite  # type: ignore
        }
        dimension_names: set[str] = set()
        for dimensions_for_group in result.grouped_by_dimensions.keys():
            dimension_names.update(dimensions_for_group.names)
        dimensions = builder.universe.conform(dimension_names)
        builder.log.debug("Building query for data IDs.")
        result.query_args = {
            "dimensions": dimensions,
            "where": builder.where,
            "dataId": result.subgraph.data_id,
            "bind": builder.bind,
        }
        if builder.dataset_query_constraint == DatasetQueryConstraintVariant.ALL:
            builder.log.debug("Constraining graph query using all datasets not marked as deferred.")
            result.query_args["datasets"] = {
                name
                for name, dataset_type_node in result.overall_inputs.items()
                if (
                    dataset_type_node.is_initial_query_constraint
                    and name not in result.empty_dimensions_dataset_types
                )
            }
            result.query_args["collections"] = builder.input_collections
        elif builder.dataset_query_constraint == DatasetQueryConstraintVariant.OFF:
            builder.log.debug("Not using dataset existence to constrain query.")
        elif builder.dataset_query_constraint == DatasetQueryConstraintVariant.LIST:
            constraint = set(builder.dataset_query_constraint)
            inputs = result.overall_inputs - result.empty_dimensions_dataset_types.keys()
            if remainder := constraint.difference(inputs):
                raise QuantumGraphBuilderError(
                    f"{remainder} dataset type(s) specified as a graph constraint, but"
                    f" do not appear as an overall input to the specified pipeline: {inputs}."
                    " Note that component datasets are not permitted as constraints."
                )
            builder.log.debug(f"Constraining graph query using {constraint}")
            result.query_args["datasets"] = constraint
            result.query_args["collections"] = builder.input_collections
        else:
            raise QuantumGraphBuilderError(
                f"Unable to handle type {builder.dataset_query_constraint} "
                "given as datasetQueryConstraint."
            )
        builder.log.verbose("Querying for data IDs with arguments:")
        builder.log.verbose("  dimensions=%s,", list(result.query_args["dimensions"].names))
        builder.log.verbose("  dataId=%s,", dict(result.query_args["dataId"].required))
        if result.query_args["where"]:
            builder.log.verbose("  where=%s,", repr(result.query_args["where"]))
        if "datasets" in result.query_args:
            builder.log.verbose("  datasets=%s,", list(result.query_args["datasets"]))
        if "collections" in result.query_args:
            builder.log.verbose("  collections=%s,", list(result.query_args["collections"]))
        with builder.butler.registry.caching_context():
            with builder.butler.registry.queryDataIds(**result.query_args).materialize() as common_data_ids:
                builder.log.debug("Expanding data IDs.")
                result.common_data_ids = common_data_ids.expanded()
                yield result

    def log_failure(self, log: LsstLogAdapter) -> None:
        """Emit an ERROR-level log message that attempts to explain
        why the initial data ID query returned no rows.

        Parameters
        ----------
        log : `logging.Logger`
            The logger to use to emit log messages.
        """
        # A single multiline log plays better with log aggregators like Loki.
        buffer = StringIO()
        try:
            buffer.write("Initial data ID query returned no rows, so QuantumGraph will be empty.\n")
            for message in self.common_data_ids.explain_no_results():
                buffer.write(message)
                buffer.write("\n")
            buffer.write(
                "To reproduce this query for debugging purposes, run "
                "Registry.queryDataIds with these arguments:\n"
            )
            # We could just repr() the queryArgs dict to get something
            # the user could make sense of, but it's friendlier to
            # put these args in an easier-to-reconstruct equivalent form
            # so they can read it more easily and copy and paste into
            # a Python terminal.
            buffer.write(f"  dimensions={list(self.query_args['dimensions'].names)},\n")
            buffer.write(f"  dataId={dict(self.query_args['dataId'].required)},\n")
            if self.query_args["where"]:
                buffer.write(f"  where={repr(self.query_args['where'])},\n")
            if "datasets" in self.query_args:
                buffer.write(f"  datasets={list(self.query_args['datasets'])},\n")
            if "collections" in self.query_args:
                buffer.write(f"  collections={list(self.query_args['collections'])},\n")
        finally:
            # If an exception was raised, write a partial.
            log.error(buffer.getvalue())
            buffer.close()



#
# LSST Data Management System
# Copyright 2008, 2009, 2010 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#


import unittest

import astropy
import lsst.daf.butler as dafButler
import lsst.pipe.base as pipeBase
import lsst.sphgeom as sphgeom
import lsst.utils.tests
import pydantic


class RegionTimeInfoTestCase(unittest.TestCase):
    """Unit tests for lsst.pipe.base.utils.RegionTimeInfo.

    Since RegionTimeInfo is a passive container that exists only for
    serialization convenience, the tests only cover serialization and assume
    e.g. input validation is handled by the component types.
    """

    def setUp(self):
        self.region = sphgeom.Circle(
            sphgeom.UnitVector3d(sphgeom.Angle.fromDegrees(34.5), sphgeom.Angle.fromDegrees(-42.0)),
            sphgeom.Angle.fromDegrees(1.0),
        )
        self.times = dafButler.Timespan(
            begin=astropy.time.Time("2013-06-17 13:34:45.775000", scale="tai", format="iso"),
            end=astropy.time.Time("2013-06-17 13:35:17.947000", scale="tai", format="iso"),
        )

    def test_init(self):
        # Both parameters are mandatory
        with self.assertRaises(ValueError):
            pipeBase.utils.RegionTimeInfo()
        with self.assertRaises(ValueError):
            pipeBase.utils.RegionTimeInfo(region=self.region)
        with self.assertRaises(ValueError):
            pipeBase.utils.RegionTimeInfo(timespan=self.times)

    def test_serialization(self):
        original = pipeBase.utils.RegionTimeInfo(region=self.region, timespan=self.times)
        adapter = pydantic.TypeAdapter(pipeBase.utils.RegionTimeInfo)

        roundtripped = adapter.validate_json(adapter.dump_json(original))
        self.assertEqual(roundtripped, original)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("Layout", "ColumnSelector", "LayoutRow")

import dataclasses
from collections.abc import Iterable, Iterator, Mapping, Set
from typing import Generic, TextIO, TypeVar

import networkx
import networkx.algorithms.components
import networkx.algorithms.dag
import networkx.algorithms.shortest_paths
import networkx.algorithms.traversal

_K = TypeVar("_K")


class Layout(Generic[_K]):
    """A class that positions nodes and edges in text-art graph visualizations.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        NetworkX export of a `.PipelineGraph` being visualized.
    column_selector : `ColumnSelector`, optional
        Parameterized helper for selecting which column each node should be
        added to.
    """

    def __init__(
        self,
        xgraph: networkx.DiGraph | networkx.MultiDiGraph,
        column_selector: ColumnSelector | None = None,
    ):
        if column_selector is None:
            column_selector = ColumnSelector()
        self._xgraph = xgraph
        self._column_selector = column_selector
        # Mapping from the column (i.e. 'x') of an already-positioned node to
        # the node keys its outgoing edges terminate at. These and all other
        # column/x variables are multiples of 2 when they refer to
        # already-existing columns, allowing potential insertion of new columns
        # between them to be represented by odd integers. Positions are also
        # inverted from the order they are usually displayed; it's best to
        # think of them as the distance (to the left) from the column where
        # text appears on the right.  This is natural because prefer for nodes
        # to be close to that text when possible (or maybe it's historical, and
        # it's just a lot of work to re-invert the algorithm now that it's
        # written).
        self._active_columns: dict[int, set[_K]] = {}
        # Mapping from node key to its column.
        self._locations: dict[_K, int] = {}
        # Minimum and maximum column (may go negative; will be shifted as
        # needed before actual display).
        self._x_min = 0
        self._x_max = 0
        # Run the algorithm!
        self._add_graph(self._xgraph)
        del self._active_columns

    def _add_graph(self, xgraph: networkx.DiGraph | networkx.MultiDiGraph) -> None:
        """Highest-level routine for the layout algorithm.

        Parameters
        ----------
        xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
            Graph or subgraph to add to the layout.
        """
        # Start by identifying unconnected subgraphs ("components"); we'll
        # display these in series, as the first of our many attempts to
        # minimize tangles of edges.
        component_xgraphs_and_orders = []
        single_nodes = []
        for component_nodes in networkx.components.weakly_connected_components(xgraph):
            if len(component_nodes) == 1:
                single_nodes.append(component_nodes.pop())
            else:
                component_xgraph = xgraph.subgraph(component_nodes)
                component_order = list(
                    networkx.algorithms.dag.lexicographical_topological_sort(component_xgraph, key=str)
                )
                component_xgraphs_and_orders.append((component_xgraph, component_order))
        # Add all single-node components in lexicographical order.
        single_nodes.sort(key=str)
        for node in single_nodes:
            self._add_single_node(node)
        # Sort component graphs by their size and then str of their first node.
        component_xgraphs_and_orders.sort(key=lambda t: (len(t[1]), str(t[1][0])))
        # Add subgraphs in that order.
        for component_xgraph, component_order in component_xgraphs_and_orders:
            self._add_connected_graph(component_xgraph, component_order)

    def _add_single_node(self, node: _K) -> None:
        """Add a single node to the layout."""
        assert node not in self._locations
        if not self._locations:
            # Special-case the first node in a component (disconnectd
            # subgraph).
            self._locations[node] = 0
            self._active_columns[0] = set(self._xgraph.successors(node))
            return
        # The candidate_x list holds columns where we could insert this node.
        # We start with new columns on the outside and a new column between
        # each pair of existing columns.  These inner nodes are usually not
        # very good candidates, but it's simplest to always include them and
        # let the penalty system in ColumnSelector take care of it.
        candidate_x = [self._x_max + 2, self._x_min - 2]
        # The columns holding edges that will connect to this node.
        connecting_x = []
        # Iterate over active columns to populate the above.
        for active_column_x, active_column_endpoints in self._active_columns.items():
            if node in active_column_endpoints:
                connecting_x.append(active_column_x)
        # Delete this node from the active columns it is in, and delete any
        # entries that now have empty sets.
        for x in connecting_x:
            destinations = self._active_columns[x]
            destinations.remove(node)
            if not destinations:
                del self._active_columns[x]
        # Add all empty columns between the current min and max as candidates.
        for x in range(self._x_min, self._x_max + 2, 2):
            if x not in self._active_columns:
                candidate_x.append(x)
        # Sort the list of connecting columns so we can easily get min and max.
        connecting_x.sort()
        best_x = min(
            candidate_x,
            key=lambda x: self._column_selector(
                connecting_x, x, self._active_columns, self._x_min, self._x_max
            ),
        )
        if best_x % 2:
            # We're inserting a new column between two existing ones; shift
            # all existing column values above this one to make room while
            # using only even numbers.
            best_x = self._shift(best_x)
        self._x_min = min(self._x_min, best_x)
        self._x_max = max(self._x_max, best_x)

        self._locations[node] = best_x
        successors = set(self._xgraph.successors(node))
        if successors:
            self._active_columns[best_x] = successors

    def _shift(self, x: int) -> int:
        """Shift all columns above the given one up by two, allowing a new
        column to be inserted while leaving all columns as even integers.
        """
        for node, old_x in self._locations.items():
            if old_x > x:
                self._locations[node] += 2
        self._active_columns = {
            old_x + 2 if old_x > x else old_x: destinations
            for old_x, destinations in self._active_columns.items()
        }
        self._x_max += 2
        return x + 1

    def _add_connected_graph(
        self, xgraph: networkx.DiGraph | networkx.MultiDiGraph, order: list[_K] | None = None
    ) -> None:
        """Add a subgraph whose nodes are connected.

        Parameters
        ----------
        xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
            Graph or subgraph to add to the layout.
        order : `list`, optional
            List providing a lexicographical/topological sort of ``xgraph``.
            Will be computed if not provided.
        """
        if order is None:
            order = list(networkx.algorithms.dag.lexicographical_topological_sort(xgraph, key=str))
        # Find the longest path between two nodes, which we'll call the
        # "backbone" of our layout; we'll step through this path and add
        # recurse via calls to `_add_graph` on the nodes that we think should
        # go between the backbone nodes.
        backbone: list[_K] = networkx.algorithms.dag.dag_longest_path(xgraph, topo_order=order)
        # Add the first backbone node and any ancestors according to the full
        # graph (it can't have ancestors in this _subgraph_ because they'd have
        # been part of the longest path themselves, but the subgraph doesn't
        # have a complete picture).
        current = backbone.pop(0)
        self._add_blockers_of(current)
        self._add_single_node(current)
        # Remember all recursive descendants of the current backbone node for
        # later use.
        descendants = frozenset(networkx.algorithms.dag.descendants(xgraph, current))
        while backbone:
            current = backbone.pop(0)
            # Find descendants of the previous node that are:
            # - in this subgraph
            # - not descendants of the current node.
            followers_of_previous = set(descendants)
            descendants = frozenset(networkx.algorithms.dag.descendants(xgraph, current))
            followers_of_previous.remove(current)
            followers_of_previous.difference_update(descendants)
            followers_of_previous.difference_update(self._locations.keys())
            # Add those followers of the previous node.  We like adding these
            # here because this terminates edges as soon as we can, freeing up
            # those columns for new new nodes.
            self._add_graph(xgraph.subgraph(followers_of_previous))
            # Add the current backbone node and all remaining blockers (which
            # may not even be in this subgraph).
            self._add_blockers_of(current)
            self._add_single_node(current)
        # Any remaining subgraph nodes were not directly connected to the
        # backbone nodes.
        remaining = xgraph.copy()
        remaining.remove_nodes_from(self._locations.keys())
        self._add_graph(remaining)

    def _add_blockers_of(self, node: _K) -> None:
        """Add all nodes that are ancestors of the given node according to the
        full graph.
        """
        blockers = set(networkx.algorithms.dag.ancestors(self._xgraph, node))
        blockers.difference_update(self._locations.keys())
        self._add_graph(self._xgraph.subgraph(blockers))

    @property
    def width(self) -> int:
        """The number of actual (not multiple-of-two) columns in the layout."""
        return (self._x_max - self._x_min) // 2

    @property
    def nodes(self) -> Iterable[_K]:
        """The graph nodes in the order they appear in the layout."""
        return self._locations.keys()

    def print(self, stream: TextIO) -> None:
        """Print the nodes (but not their edges) as symbols in the right
        locations.

        This is intended for use as a debugging diagnostic, not part of a real
        visualization system.

        Parameters
        ----------
        stream : `io.TextIO`
            Output stream to use for printing.
        """
        for row in self:
            print(f"{' ' * row.x}●{' ' * (self.width - row.x)} {row.node}", file=stream)

    def _external_location(self, x: int) -> int:
        """Return the actual (not multiple-of-two, stating from zero) location,
        given the internal multiple-of-two.
        """
        return (self._x_max - x) // 2

    def __iter__(self) -> Iterator[LayoutRow]:
        active_edges: dict[_K, set[_K]] = {}
        for node, node_x in self._locations.items():
            row = LayoutRow(node, self._external_location(node_x))
            for origin, destinations in active_edges.items():
                if node in destinations:
                    row.connecting.append((self._external_location(self._locations[origin]), origin))
                    destinations.remove(node)
                if destinations:
                    row.continuing.append(
                        (self._external_location(self._locations[origin]), origin, frozenset(destinations))
                    )
            row.connecting.sort(key=str)
            row.continuing.sort(key=str)
            yield row
            active_edges[node] = set(self._xgraph.successors(node))


@dataclasses.dataclass
class LayoutRow(Generic[_K]):
    """Information about a single text-art row in a graph."""

    node: _K
    """Key for the node in the exported NetworkX graph."""

    x: int
    """Column of the node's symbol and its outgoing edges."""

    connecting: list[tuple[int, _K]] = dataclasses.field(default_factory=list)
    """The columns and node keys of edges that terminate at this row.
    """

    continuing: list[tuple[int, _K, frozenset[_K]]] = dataclasses.field(default_factory=list)
    """The columns and node keys of edges that continue through this row.
    """


@dataclasses.dataclass
class ColumnSelector:
    """Helper class that weighs the columns a new node could be added to in a
    text DAG visualization.
    """

    crossing_penalty: int = 1
    """Penalty for each ongoing edge the new node's outgoing edge would have to
    "hop" if it were put at the candidate column.
    """

    interior_penalty: int = 1
    """Penalty for adding a new column to the layout between existing columns
    (in addition to `insertion_penaly`.
    """

    insertion_penalty: int = 2
    """Penalty for adding a new column to the layout instead of reusing an
    empty one.

    This penalty is applied even when there is no empty column; it just cancels
    out in that case because it's applied to all candidate columns.
    """

    def __call__(
        self,
        connecting_x: list[int],
        node_x: int,
        active_columns: Mapping[int, Set[_K]],
        x_min: int,
        x_max: int,
    ) -> int:
        """Compute the penalty score for adding a node in the given column.

        Parameters
        ----------
        connecting_x : `list` [ `int` ]
            The columns of incoming edges for this node.  All values are even.
        node_x : `int`
            The column being considered for the new node.  Will be odd if it
            proposes an insertion between existing columns, or outside the
            bounds of ``x_min`` and ``x_max`` if it proposes an insertion
            on a side.
        active_columns : `~collections.abc.Mapping` [ `int`, \
                `~collections.abc.Set` ]
            The columns of nodes already in the visualization (in previous
            lines) and the nodes at which their edges terminate.  All keys are
            even.
        x_min : `int`
            Current minimum column position (inclusive).  Always even.
        x_max : `int`
            Current maximum column position (exclusive).  Always even.

        Returns
        -------
        penalty : `int`
            Penalty score for this location.  Nodes should be placed at the
            column with the lowest penalty.
        """
        # Start with a penalty for inserting a new column between two existing
        # columns or on either side, if that's what this is (i.e. x is odd).
        penalty = (node_x % 2) * (self.interior_penalty + self.insertion_penalty)
        if node_x < x_min:
            penalty += self.insertion_penalty
        elif node_x > x_max:
            penalty += self.insertion_penalty
        # If there are no active edges connecting to this node, we're done.
        if not connecting_x:
            return penalty
        # Find the bounds of the horizontal lines that connect
        horizontal_min_x = min(connecting_x[0], node_x)
        horizontal_max_x = max(connecting_x[-1], node_x)
        # Add the (scaled) number of unrelated continuing (vertical) edges that
        # the (horizontal) input edges for this node would have to "hop".
        penalty += sum(
            self.crossing_penalty
            for x in range(horizontal_min_x, horizontal_max_x + 2)
            if x in active_columns and x not in connecting_x
        )
        return penalty
